package main

import (
	"context"
	"database/sql"
	"fmt"
	"log"
	"sync"
	"time"

	_ "github.com/mattn/go-sqlite3"
)

// workは`d`で指定された時間ブロックし、なにかの作業をしている
// 状態を模擬的に再現している
func work(ctx context.Context, d time.Duration) error {
	select {
	case <-ctx.Done():
		// 呼び出し側から渡されたcontext.Contextオブジェクトが
		// キャンセルされたら、以下のtime.Afterを待たずにこの
		// 関数は終了する
		fmt.Println("canceled")
		return ctx.Err()
	case <-time.After(d):
		// 指定された時間がたった後、nilを返す
		fmt.Println("fired")
		return nil
	}
}

// 最も基本的なcontextの呼び出し。親contextは特にキャンセルされず、
// workは5秒処理を続け、終了する
func ExampleContext() {
	ctx, cancel := context.WithCancel(context.Background())
	defer cancel()

	work(ctx, 5*time.Second)
	// OUTPUT:
	// fired
}

// 呼び出し側が1秒でcontextをキャンセルするので、work関数は5秒を
// 待たずにキャンセルされる
func ExampleTimeout() {
	ctx, cancel := context.WithTimeout(context.Background(), time.Second)
	defer cancel()

	work(ctx, 5*time.Second)
	// OUTPUT:
	// canceled
}

func ExampleDatabaseRollback() {
	ctx, cancel := context.WithCancel(context.Background())
	defer cancel()

	db, err := sql.Open("sqlite3", ":memory:")
	if err != nil {
		log.Printf("failed to open database: %s", err)
		return
	}

	// トランザクションの開始
	tx, err := db.BeginTx(ctx, nil)
	if err != nil {
		log.Printf("failed to start transaction: %s", err)
		return
	}

	// これまではどこかのタイミングで tx.Rollback()を呼ばないと
	// いけなかったが、tx.Commit()のあとで呼ぶと色々問題があった。
	// BeginTxとcontext.Contextを使う事によって、この関数を抜ける
	// タイミングでcancelが呼ばれ、トランザクションもキャンセルされる

	// これはコンパイルを通すためのダミーコード
	_, _ = tx.Query("SELECT 1")
}

type legacyLoop struct {
	cond *sync.Cond
	done bool
	loop bool
	mu   sync.RWMutex
}

func NewLegacyLoop() *legacyLoop {
	return &legacyLoop{
		cond: sync.NewCond(&sync.Mutex{}),
		loop: false,
	}
}

func (l *legacyLoop) Start() {
	// loop はchan でも代用できるが、Start/Stopが複数回呼ばれても
	// 良いようにするには初期化のタイミングが重要になってくるため
	// ここではboolのほうが安全だと判断した
	l.mu.Lock()
	l.loop = true
	l.mu.Unlock()

	l.cond.L.Lock()
	l.done = false
	l.cond.L.Unlock()

	go func() {
		defer func() {
			l.cond.L.Lock()
			l.done = true
			l.cond.Broadcast()
			l.cond.L.Unlock()
		}()

		for { // 無限ループ
			if !l.running() {
				return
			}

			// 通常の処理（ここではログ出力とSleepのみ）
			log.Printf("hello")
			time.Sleep(time.Second)
		}
	}()
}

func (l *legacyLoop) running() bool {
	l.mu.RLock()
	defer l.mu.RUnlock()
	return l.loop
}

func (l *legacyLoop) Stop() {
	l.mu.Lock()
	defer l.mu.Unlock()
	l.loop = false
}

func (l *legacyLoop) Wait() {
	l.cond.L.Lock()
	for !l.done {
		log.Printf("wait...")
		l.cond.Wait()
	}
	l.cond.L.Unlock()
}

// 「厳密」に処理した場合のStart/Stopを使ったループ機構
// ただし、これでもStart/Stopがあるタイミングで複数回同時に
// 呼ばれるとおかしな事になる可能性がある
//
// ポイントとしては
// (1) すでにStart()とStop()というAPIを切り出してしまった以上、
//     これを間違ったタイミングで呼ばせないようにするには、
//     内部でそれなりの工夫をしないといけない、ということ
// (2) 非同期にgoroutineとの間でやりとりをしないといけないので
//     ロックでデータの整合性を保つ必要があるということ
func ExampleLegacyStartStopLoop() {
	l := NewLegacyLoop()
	l.Start()

	time.Sleep(5 * time.Second)
	// 5秒後にキャンセル
	l.Stop()
	l.Wait()
}

type contextLoop struct {
	// ループの制御に特に何もいらない
}

// ループ用のコード
func (o *contextLoop) Run(ctx context.Context) {
	for { // 無限ループ
		select {
		case <-ctx.Done():
			// 呼び出し側でcontextをキャンセルしたらこのメソッドを
			// 抜ける（＝ループ終了）
			return
		}

		// 通常の処理（ここではログ出力とSleepのみ）
		log.Printf("hello")
		time.Sleep(time.Second)
	}
}

// context.Contextを使った場合のループ
// ctx.Done()以外気にする必要がないので、非常にシンプルなものになる
func ExampleContextLoop() {
	ctx, cancel := context.WithCancel(context.Background())
	defer cancel()

	// 特に初期化もいらないのでゼロ値を利用できる
	var l contextLoop

	// goroutineを明示的に作成
	go l.Run(ctx)

	time.Sleep(5 * time.Second)

	// 5秒後にループを抜ける際に自動的にcontextもキャンセルされ
	// 別goroutineのコードもキャンセルされる
}

// context.Contextを使った場合のループ
// 本文のようにcontextに5秒のタイムアウトを設定した場合
func ExampleContextLoopTimeout() {
	// 下記contextは5秒後に自動的にキャンセルされる
	ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
	defer cancel()

	// 特に初期化もいらないのでゼロ値を利用できる
	var l contextLoop

	// goroutineを明示的に作成
	go l.Run(ctx)

	// 他に待つ条件もないので、ここではcontextが終わるのを待つ事で
	// お茶を濁す。下記条件が満たされてこの関数が終了すると、同時に
	// 上記goroutineもキレイに終了する
	<-ctx.Done()
}
